PROGRAM mprobs;

{$E+,N+}  { If an 8087 is present, use it, otherwise use the 8087 Emulator }

USES  DOS;

{   MONKEY PROBABILITIES  Version 1.0

    This Turbo Pascal program calculates the probabilities encountered by
the standard "Alphabet" test in MONKEY.PAS, which uses cumulative selection
to produce a target string (e.g. the alphabet in alphabetical order) through
random changes.  The calculations performed herein use Markov chains
implemented through stochastic matrices.  The resultant probability vectors
from each step in the Markov chain are written out to the text file,
MPROBS.LST.

    To use this program, set the desired values of the constants CHAR_SIZE
(size of the character set -- normally 26), MSG_SIZE (length of the target
string -- 26 in the Alphabet test), MAX_TRIES (total number of iterations
calculated -- nominally 100), and MEMBERS (number of copies per iteration --
in effect, the generation size).  Of the above constants, MEMBERS is most
likely to be varied in order to see what effect the generation size has on
the probabilities.
    Once the constants have been set, then compile and run the program.  The
program will run through the indicated number of iterations, outputting the
results to the file, MPROBS.LST, and terminate upon completion.  The program
displays the current iteration number on the screen to indicate that it is
still running.

    This program was written in Turbo Pascal v5.0 and requires the Extended
Real data type (3.4 E -4932 .. 1.1 E 4932).  By using the $E+ and $N+ compiler
directives, this program will automatically detect the presence of a math
co-processor and use it if it exists or use software emulation if it doesn't.
If you use Turbo Pascal v4.0 or another Pascal compiler, you may need to modify
this program to be able to run it.

    Since there is a lot of number-crunching involved in the matrix operations,
this program is SLOW!  Representative run-times for the standard parameters
(CHAR_SIZE = 26, MSG_SIZE = 26, MAX_TRIES = 100, any value for MEMBERS) are:

             on an 8-MHz XT with 8087    -- 19 min,
         and on an 8-MHz XT without 8087 -- 58 min.


    For more information on cumulative selection, read the text files,
MONKEY.DOC and MPROBS.DOC, the source code of MONKEY.PAS, and Chapter 3 of
Richard Dawkins' book, _The Blind Watchmaker_.


    Listed below are short descriptions of the parameters and probabilities
used:

c = size of character set = 26
k = count of correct letters
n = target size = 26
m = number of members per generation

P1 = prob to generate the right letter in a given position = 1/c
Q1 = prob to generate the wrong letter in a given position = 1 - P1 = (c-1)/c
P2 = prob to select position of a wrong letter = (n-k)/n
Q2 = prob to select position of a correct letter = k/n

P3 = single prob to make next step = P1 * P2 = (n-k)/nc
Q3 = single prob not to make step  = 1 - P3 = (nc - n + k)/nc
P4 = single prob to slip back      = Q1 * Q2 = k*(c-1)/nc
Q4 = single prob to not slip back  = 1 - P4 = (nc - kc + k)/nc
P5 = single prob of no change      = 1 - (P3 + P4) = Q3 * Q4  (?)
                                   = (nc - n + 2k - kc)/nc

P6 = joint prob to make next step = OR(i in 1..m,P3i) = 1 - Q3^m
P7 = joint prob to slip back      = AND(i in 1..m,P4i) = P4^m
P8 = joint prob of no change      = 1 - (P6 + P7) = Q3^m - P4^m

}


(* CHANGE THE FOLLOWING CONSTANTS TO GENERATE PROBABILITIES
                 FOR DIFFERENT CASES  *)

TYPE
  test_types = (individual,population);

CONST
  test_type = POPULATION;
  char_size = 26;             { Size of character set }
  msg_size = 26;              { Length of target string (maximum of 50) }
  nc = msg_size * char_size;  { Pre-calculate often-used term }
  members = 100;              { Define number of copies per iteration }
  max_tries = 100;            { Set number of steps to be taken }


TYPE
  real_type = Extended;               { 8087 type:  3.4E-4932 .. 1.1E4932 }
  prob_vector = ARRAY[0..msg_size] OF real_type;    { Vector of probabilities}
  prob_array = ARRAY[0..msg_size,0..msg_size] OF real_type;
                                                    { Matrix of probabilities}

  prob_rec = RECORD                   { Set of up to ten probabilities (p) }
    p, q   :ARRAY[1..10] OF real_type;  {  and their inverses (q) for a    }
  END;                                {  given node of the state machine   }

  time_rec = RECORD                   { Time Record Type }
    hour,
    min,
    sec,
    sec100  :INTEGER;
  END;


VAR
  f         :TEXT;                    { Report Output file }
  prob      :ARRAY[0..msg_size] OF prob_rec;
                                      { Probabilities for each state }
  init_array,                         { Initial Transition Matrix }
  markov    :prob_array;              { Markov chain }
  init_cond :prob_vector;             { Vector of initial probabilities }

  start_time,                         { Start and Stop times and duration  }
  stop_time,                          {   for measuring elapsed time       }
  duration   :time_rec;



(* MATH SUPPORT *)

{ Constants and Variables for LOG and POWER }

CONST
  max_val = 1.0E30;         { Within range of REAL type }
  min_val = 1.0E-30;
  big_e = max_val;
  little_e = min_val;

VAR
  big_ln,                   { LN(max_val)  }
  little_ln  :REAL;         { LN(min_val)  }


FUNCTION log(v:real_type):real_type;

{  LOG returns the common log (base 10) of the value, V.  It works for the
   defined REAL_TYPE (Real, Double, or Extended).  The extra testing and
   processing is to get around the range limitations of the LN function. }

VAR
  i,
  count :INTEGER;
  e,
  ln_e,
  x     :real_type;

BEGIN
  IF v = 0.0
    THEN
      x := -max_val
    ELSE
      BEGIN
      IF v > max_val
        THEN
          BEGIN
          e := big_e;
          END
        ELSE
          IF v < min_val
            THEN
              BEGIN
              e := little_e;
              END
            ELSE
              e := 1.0;
      ln_e := LN(e);
      count := 0;
      WHILE (v < min_val) OR (v > max_val) DO
        BEGIN
        INC(count);
        v := v / e;
        END;
      x := LN(v);
      FOR i:=1 TO count DO
        x := x + ln_e;
      END;
  log := x;
END;


FUNCTION power(v:real_type;expon:INTEGER):real_type;

{  POWER takes the value, V, and raises it to the EXPON power, the value of
   which it returns.  It works for the defined REAL_TYPE (Real, Double, or
   Extended).  The extra testing and processing is to get around the range
   limitations of the LN and the EXP function. }

VAR
  i,
  count :INTEGER;
  e,
  ln_e,
  x     :real_type;

BEGIN
  IF v = 0.0
    THEN
      x := 0.0
    ELSE
      BEGIN
      x := expon * log(v);
      IF x > big_ln
        THEN
          BEGIN
          e := big_e;
          END
        ELSE
          IF x < little_ln
            THEN
              BEGIN
              e := little_e;
              END
            ELSE
              e := 1.0;
      ln_e := LN(e);
      count := 0;
      WHILE (x < little_ln) OR (x > big_ln) DO
        BEGIN
        INC(count);
        x := x - ln_e;
        END;
      x := EXP(x);
      FOR i:=1 TO count DO
        x := x * e;
      END;
  power := x;
END;



(* PROBABILITY FUNCTIONS *)


FUNCTION inverse(x:real_type):real_type;

{  INVERSE returns the inverse of the probability, X. }

BEGIN
  inverse := 1 - x;
END;


FUNCTION anded(p:real_type;m:INTEGER):real_type;

{  ANDED calculates the probability of P occurring M times. }

BEGIN
  anded := power(p,m);
END;


FUNCTION ored(p:real_type;m:INTEGER):real_type;

{  ORED calculates the probability of P occurring at least once in M tries. }

BEGIN
  ored := inverse(anded(inverse(p),m));
END;


PROCEDURE node(k:BYTE);

{  NODE calculates the probabilities associated with the k'th state in the
   state machine (see MPROBS.DOC).  The procedure is called repeatedly for all
   states, the state number being passed in as K.  See MPROBS.DOC for a more
   complete description of each probability.                               }

BEGIN
  WITH prob[k] DO
    BEGIN
    p[1] := 1.0 / char_size;    { the right letter in a given position }
    q[1] :=  1 - p[1];          { the right letter in a given position }
    q[2] := k / msg_size;       { select position of a wrong letter }
    p[2] :=  1 - q[2];          { select position of a correct letter }
    p[3] := p[1] * p[2];        { single prob to make next step }
    q[3] := 1 - p[3];           { single prob not to make next step }
    p[4] := q[1] * q[2];        { single prob to slip back }
    q[4] :=  1 - p[4];          { single prob not to slip back }
    p[5] := 1 - (p[3] + p[4]);  { single prob of no change }
    q[5] :=  1 - p[5];          { single prob of some change }
    IF k = msg_size
      THEN
        BEGIN           { In the target state, there is no probability }
        p[6] := 0.0;    {  of either advancing (p[6]) or slipping back p[7],}
        q[6] := 1.0;    {  only a dead certainty of remaining.   }
        p[7] := 0.0;
        q[7] := 1.0;
        q[8] := 0.0;
        p[8] := 1.0;
        END
      ELSE
        BEGIN
        p[6] := ored(p[3],members);   { joint prob to make next step }
        q[6] := 1 - p[6];             { joint prob not to make next step }
        p[7] := anded(p[4],members);  { joint prob to slip back }
        q[7] := 1 - p[7];             { joint prob not to slip back }
        q[8] := p[6] + p[7];          { joint prob of some change }
        p[8] := 1 - q[8];             { joint prob of no change }
        END;
    END;
END;


PROCEDURE build_chain(k:BYTE);

{  BUILD_CHAIN constructs the k'th row of the Markovian transition matrix
   (see MPROBS.DOC).  The procedure is called repeatedly for all states,
   the state number being passed in as K.  All entries are zero except for
   the (k-1)'th, the k'th, and the (k+1)'th columns, which contain the
   probabilities of transitioning to those states.
   Of course, the 0'th state has no (-1)'th transition and the final, 26'th,
   state only transitions to itself (this known as the absorbing state).
   Both of these special cases are handled implicitly.                  }

VAR
  i, j  : BYTE;

BEGIN
  FOR i:=0 TO msg_size DO
    CASE test_type OF
      individual:
          IF i = (k - 1)
            THEN  init_array[k,i] := prob[k].p[4]
            ELSE
              IF i = k
                THEN  init_array[k,i] := prob[k].p[5]
                ELSE
                  IF i = (k+1)
                    THEN  init_array[k,i] := prob[k].p[3]
                    ELSE  init_array[k,i] := 0.0;
      population:
          IF i = (k - 1)
            THEN  init_array[k,i] := prob[k].p[7]
            ELSE
              IF i = k
                THEN  init_array[k,i] := prob[k].p[8]
                ELSE
                  IF i = (k+1)
                    THEN  init_array[k,i] := prob[k].p[6]
                    ELSE  init_array[k,i] := 0.0;
    END;  {CASE}
END;


PROCEDURE vmult(VAR v,c:prob_vector;VAR a:prob_array);

{  VMULT performs a vector-to-matrix multiplication, which yields a vector.
   I.e., it multiples the vector, V, to the matrix, A, and returns the
   resultant vector, C.                                                }

VAR
  i, j :BYTE;

BEGIN
  FOR i:=0 TO msg_size DO
    BEGIN
    c[i] := 0.0;
    FOR j:=0 TO msg_size DO
      c[i] := c[i] + v[j] * a[j,i];
    END;
END;


PROCEDURE mmult(VAR a,b:prob_array);

{  MMULT performs a matrix-to-matrix multiplication, which yields a matrix.
   I.e., it multiples the matrix, A, to the matrix, B, (AB).  The resultant
   matrix, C, is handled dynamically and, after the multiplication is
   complete, *is disposed of after its values are passed back in A.      }

VAR
  c  :^prob_array;
  i, j, k :BYTE;

BEGIN
  NEW(c);
  FOR i:=0 TO msg_size DO
    FOR j:=0 TO msg_size DO
      BEGIN
      c^[i,j] := 0.0;
      FOR k:=0 TO msg_size DO
        c^[i,j] := c^[i,j] + a[i,k] * b[k,j];
      END;
  a := c^;
  DISPOSE(c);
END;


PROCEDURE get_elapsed_time(start:time_rec;VAR duration:time_rec);

{  GET_ELAPSED_TIME reads the stop time and calculates the elapsed time in
   hours, minutes, seconds, and 1/100's of seconds.  }

  PROCEDURE adjust_time(VAR hi_unit,lo_unit:INTEGER;factor:INTEGER);

  BEGIN
    IF lo_unit < 0
      THEN
        BEGIN
        INC(lo_unit,factor);
        DEC(hi_unit);
        END;
  END;

BEGIN
  WITH stop_time DO
    GetTime(WORD(hour),WORD(min),WORD(sec),WORD(sec100));
  WITH duration DO
    BEGIN
    hour := stop_time.hour - start.hour;
    min := stop_time.min - start.min;
    sec := stop_time.sec - start.sec;
    sec100 := stop_time.sec100 - start.sec100;
    adjust_time(sec,sec100,100);
    adjust_time(min,sec,60);
    adjust_time(hour,min,60);
    END;
END;


PROCEDURE report_time(duration:time_rec);

{  REPORT_TIME displays the amount of time the program had run.  }

BEGIN
  WRITELN;
  WITH duration DO
    BEGIN
    WRITE('Elapsed Time = ',hour:2,':');
    IF min < 10
      THEN  WRITE('0',min:1,':')
      ELSE  WRITE(min:2,':');
    IF sec < 10
      THEN  WRITE('0',sec:1,'.')
      ELSE  WRITE(sec:2,'.');
    IF sec100 < 10
      THEN  WRITE('0',sec100:1)
      ELSE  WRITE(sec100:2);
    WRITELN;
    END;
END;


PROCEDURE do_report;

{  DO_REPORT performs the calls to construct the Markov chain for each step
   and outputs the resultant vector of probabilities to the text file,
   MPROBS.LST.  See MPROBS.DOC for more information.
   As the chain is being constructed, the step number is displayed on the
   screen to inform the user of how far the program has gotten.  Also, the
   amount of time that it took to construct the chain is displayed on the
   screen at the end of the calculations.  (Hint:  it takes a long time) }

VAR
  result :prob_vector;
  sum    :real_type;
  i, j   :BYTE;

BEGIN
  CASE test_type OF
    individual :  ASSIGN(f,'MPROBS_I.LST');
    population :  ASSIGN(f,'MPROBS_P.LST');
  END;  {CASE}

  REWRITE(f);
  WRITELN(f,'String Length = ',msg_size);
  WRITELN(f,'Character Set Size = ',char_size);
  WRITELN(f,'Generation Size = ',members);
  WRITELN(f);
  WITH start_time DO
    GetTime(WORD(hour),WORD(min),WORD(sec),WORD(sec100));
  FOR i:=2 TO max_tries DO
    BEGIN
    mmult(markov,init_array);
    vmult(init_cond,result,markov);
    WRITELN(i);
    WRITELN(f,'After ',i,' Tries:');
    sum := 0.0;
    FOR j:=0 TO msg_size DO
      BEGIN
      sum := sum + result[j];
      WRITE(f,j:6,'.  ',result[j]);
      IF ODD(j)
        THEN  WRITELN(f);
      END;
    WRITELN(f);
    WRITELN(f,'Total Prob = ',sum);
    WRITELN(f);
    END;
  CLOSE(f);
  get_elapsed_time(start_time,duration);
  report_time(duration);
END;



{MAIN}

{  MAIN initializes values and the matrices and vectors of the Markov chain.
   Then it calls DO_REPORT to process the chain and output the results to
   a disk file.     }

VAR
  i  :BYTE;

BEGIN

{  Initialize test values for LOG and POWER  }

  big_ln := LN(big_e);
  little_ln := LN(little_e);

{  Create the Transition Matrix }

  FOR i:=0 TO msg_size DO    { For each State:  }
    BEGIN
    node(i);                 {   Calculate its probabilities }
    build_chain(i);          {   Construct its row in the transition matrix }
    END;

{  Initialize the Markov chain and the initial probabilities vector }

  markov := init_array;
  FOR i:=0 TO msg_size DO
    init_cond[i] := init_array[0,i];

{  Process the Markov chain and write the results of each step to the
   output file.      }

  do_report;

END.
